// code by jph

class aniSkeleton : public animated {
  public:
    aniSkeleton(void);
    inline ~aniSkeleton(void) {};
    void control(void);
    void motion(void);            // rarely overloaded!
    void draw(void);              // mandatory
    int isConsistent(void);
    void normalize(void);
    void correct(void);           // correction of errors due to numerical integration
    void rotate(tensor<double> axs);
    tensor<animated*> whos(void); // requires canonic link structure
    tensor<double> momentum_lin(void);
    tensor<double> momentum_ang(void);
    double energy_kin(void);
    tensor<double> elem_lhs(aniLink* lnk0,int ind0,aniLink* lnk1,int ind1);
    
    linked<aniLink*> link;        // intialized upon initialization
    tensor<double> mlin_;         // 4 vector of initial total linear momentum
    tensor<double> mang_;         // 4 vector of initial total angular momentum
    double ekin_;                 // initial kinetic energy
};

aniSkeleton::aniSkeleton(void) {
  mlin_=tensor<double>(1,4);
  mang_=tensor<double>(1,4);
  ekin_=0;
  status=ANI_DRAW_DYN;
}

int aniSkeleton::isConsistent(void) { // ok
  aniLink* lnk;
  while (link.token(lnk))
    lnk->isConsistent();
  return 1;
}

void aniSkeleton::control(void) {
  int c0,c1;
  aniLink* lnk;
  while (link.token(lnk))
    lnk->control();
}

void aniSkeleton::motion(void) {
  int c0,c1;
  aniLink* lnk;
  tensor<aniLink*> arr=link.toTensor();
  tensor<tensor<double>*> lhs(2,arr.numel,arr.numel); // composition of block matrix
  tensor<tensor<double>*> rhs(1,arr.numel);           // composition of block vector
  int rank=0;  
  
  for (c0=0;c0<arr.numel;++c0) {
    tensor<double> ect(1,6);
    for (c1=0;c1<arr.numel;++c1) {
      lhs(c0,c1)=arr(c0)->control_lhs(arr(c1));
      ect+=arr(c0)->control_ect(arr(c1))*gl3v(arr(c1)->ect);
    }
    rhs(c0)=arr(c0)->control_rhs();
    *rhs(c0)-=ect;
    
    rank+=arr(c0)->rank;
  }
  tensor<double>::precision=1e-5;
  tensor<double> LHS=tensor<double>(lhs);
  tensor<double> sol=LHS.inverse(rank)*tensor<double>(rhs);  
  
  if (tensor<double>::valid!=rank) {
    printf("rank %i, sol=%s\n ",tensor<double>::valid,sol.toString("%6.3f "));
    printf("\n-----------------\n%s\n\n",LHS.toString());
    exit(1);
  }
  
  int cnt=0;
  while (link.token(lnk)) {
    lnk->intake(&sol.data[cnt]);    
    cnt+=6;
  }
  
  tensor<animated*> bdy=whos();
  for (c0=0;c0<bdy.numel;c0++)
    bdy(c0)->motion();
}

void aniSkeleton::draw(void) {
  int c0;
  aniLink* tok;
  while (link.token(tok)) // draw links
    tok->draw();
  if (status&ANI_DRAW_DYN)
    while (link.token(tok)) // draw link dynamics
      tok->draw_dyn();
  if (keyboard('s')) {
    status^=ANI_DRAW_DYN*keyboard_down('d');
  }
}

void aniSkeleton::normalize(void) { // ok
  int c0,c1;
  tensor<animated*> ani=whos();
	//center of total mass at 0	
	double fac=0;
	tensor<double> cen(1,4);
	for (c0=0;c0<ani.numel;++c0)
	  cen+=ani(c0)->pos[3]*ani(c0)->mass;
	fac=1/cen.data[3];
	cen*=fac; // inverse of total mass
	for (c0=0;c0<ani.numel;++c0)
	  for (c1=0;c1<3;c1++)
	    ani(c0)->pos.data[12+c1]-=cen.data[c1];
	// total linear momentum = 0
	cen=0;
	for (c0=0;c0<ani.numel;++c0)
	  cen+=ani(c0)->momentum_lin();
	cen*=fac; // inverse of total mass
	for (c0=0;c0<ani.numel;++c0)
    ani(c0)->vel-=cen;
  // total angular momentum = 0
  tensor<double> lhs(2,4,4);
  for (c0=0;c0<ani.numel;++c0) 
    lhs+=gl4r(ani(c0)->pos*ani(c0)->ine*glI(ani(c0)->pos)-glD(ani(c0)->pos[3])*glD(ani(c0)->pos[3])*ani(c0)->mass);
  lhs(3,3)=1;
  rotate(lhs.solve(momentum_ang()*-1.0));  
  mlin_=momentum_lin(); // should be all zeros
  mang_=momentum_ang(); // should be all zeros
  assert(mlin_==tensor<double>(1,4));
  assert(mang_==tensor<double>(1,4));
  ekin_=energy_kin();
}

void aniSkeleton::rotate(tensor<double> axs) { // sign convention of right hand thumb + fist
  int c0;
  tensor<animated*> ani=whos();
  for (c0=0;c0<ani.numel;++c0) {
    ani(c0)->vel+=axs.x(ani(c0)->pos[3]);
    ani(c0)->ome+=ani(c0)->pos.solve(axs);
  }
}

tensor<animated*> aniSkeleton::whos(void) {
  aniLink* tok;
  tensor<animated*> ret(1,1+link.numel);
  int c0=0;
  if (link.token(tok)) {
    ret(c0++)=tok->ani[0];
    ret(c0++)=tok->ani[1];
    while (link.token(tok))
      ret(c0++)=tok->ani[1];
  }
  return ret;
}

tensor<double> aniSkeleton::momentum_lin() {
  tensor<double> ret(1,4);
  tensor<animated*> ani=whos();
  for (int c0=0;c0<ani.numel;++c0)
    ret+=ani(c0)->momentum_lin();
  return ret;
}

tensor<double> aniSkeleton::momentum_ang() {
  tensor<double> ret(1,4);
  tensor<animated*> ani=whos();
  for (int c0=0;c0<ani.numel;++c0)
    ret+=ani(c0)->momentum_ang();
  return ret;
}

double aniSkeleton::energy_kin() {
  double ret=0;
  tensor<animated*> ani=whos();
  for (int c0=0;c0<ani.numel;++c0)
    ret+=ani(c0)->energy_kin();
  return ret;
}

void aniSkeleton::correct(void) {
  aniLink* lnk;
  while (link.token(lnk))
    lnk->correct();
}
